-- Filename:     seq_mult_2_algo.vhd
-- Filetype:     VHDL code (algorithmic)
-- Date:         30 march 2012
-- Update:       Changed to numeric_std for addition
-- Description:  Algorithmic description of the multiplier
-- Author:       J. op den Brouw
-- State:        demo
-- Error:        -
-- Version:      1.1
-- Copyright:    (c)2012, De Haagse Hogeschool

-- The VHDL file contains the algorithmic description
-- of the sequential multiplier version 2. It is
-- based on the well known add-and-shift school method
-- for multiplication. The pseudo code is as follows:
--

--   ready := 0; running := 0;            -- not ready
--   A := 0; B := 0; P := 0;              -- (reset)
--   forever do                           -- hardware runs continous
--     if start = 1 and running = 0 then  -- start! initialize
--       A := input-A; B := input-B; P:= 0; ready := 0; running := 1;
--     elsif B ≠ 0 and running  = 1 then  -- B still has ones ...
--       if b(0) = '1' then
--         P := P + A;
--       end if;
--       left-shift A; right-shift B;
--     else                               -- B has no ones left
--       if start = 0 then running := 0; end if;
--       if running = 1 and B = 0 then ready := 1; else ready := 0; end if;
--     end if;
--   end for;
--
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- The Sequential Mulitplier Version 2
entity seq_mult_2_algo is
	generic (n : integer := 4);
	port (clk : in std_logic;
			areset : in std_logic;
			start : in std_logic;
			dataa : in unsigned(n-1 downto 0);
			datab : in unsigned(n-1 downto 0);
			resultp : out unsigned(2*n-1 downto 0);
			ready : out std_logic
	);
end seq_mult_2_algo;

architecture algoritmic of seq_mult_2_algo is
-- Signals is for debug only, synthesizer removes signal, but complains
signal running_debug : std_logic;
signal rega_debug : unsigned (2*n-1 downto 0);
signal regb_debug : unsigned (n-1 downto 0);
signal regp_debug : unsigned (2*n-1 downto 0);
begin

	process (clk, areset) is
	variable rega : unsigned (2*n-1 downto 0);
	variable regb : unsigned (n-1 downto 0);
	variable regp : unsigned (2*n-1 downto 0);
	variable running : std_logic;
	constant n_zeros : unsigned (n-1 downto 0) := (others => '0');
	begin
		-- Asynchronous reset
		if areset = '1' then
			-- Flag stopped
			running := '0';
			rega := (others => '0');
			regb := (others => '0');
			regp := (others => '0');
		-- On positive edge
		elsif rising_edge(clk) then
			-- If start is pressed and we're not running then load or
			-- initialize registers and flag we're running!
			if start = '1' and running = '0' then
				rega := n_zeros & dataa;
				regb := datab;
				regp := (others => '0');
				running := '1';
         -- As long as B has 1's and we're running then add A to P
			-- (of course, only if b(0) is 1) and shift A en B accordingly
			elsif regb /= n_zeros and running = '1' then
				if regb(0) = '1' then
					regp := regp + rega;
				end if;
				rega := rega(2*n-2 downto 0) & '0';
				regb := '0' & regb(n-1 downto 1);
			-- B is 0 and we're running...
			elsif regb = n_zeros and running = '1' then
				-- If start is released, we're not running (stopped)
				if start = '0' then
					running := '0';
				end if;
			end if; -- clock edge
		end if; -- async reset

		-- If we're running and B is 0 then we're ready
		if running = '1' and regb = n_zeros then
			ready <= '1';
		else
			ready <= '0';
		end if;
		
		-- Output result
		resultp <= regp;

		-- Debug outputs
		running_debug <= running;
		rega_debug <= rega;
		regb_debug <= regb;
		regp_debug <= regp;
		
	end process;
end algoritmic;
